from enum import Enum, auto


class Category(Enum):
    """
        a category represents a class of object that may be contained in a ref
        categories may be hierarchical
    """
    STUDY = auto()
    AUTHOR_LIST = auto()
    AUTHOR = auto()
    AFFILIATION = auto()
    DESCRIPTOR = auto()
    KEYWORD = auto()
    PUBLICATION_TYPE = auto()
    GRANT = auto()

    def list_appendable(self) -> bool:
        """
            tag content of this category may simply be placed in a list
        """
        return self in [Category.GRANT, Category.PUBLICATION_TYPE, Category.KEYWORD, Category.AFFILIATION,
                        Category.AUTHOR_LIST, Category.DESCRIPTOR]

    def attribute_appendable(self) -> bool:
        """
            tag content of this category
        """
        return self in [Category.STUDY, Category.AUTHOR]

    def get_start_tag(self) -> str:
        """
            for tags that occur in sequence (i.e. list & attribute appendable), get the tag that marks the start of a
            new entity. will throw for Enums that aren't both list & attribute appendable
        """
        if self == Category.AUTHOR:
            return 'FAU'

        raise ValueError('Only attribute appendable categories rooted beneath a list appendable have a start tag')

    def get_category_path(self) -> list:
        """
            defines how to navigate to the location to append the next tag of the category
        """
        return _category_to_path[self]

    def get_attribute_name(self) -> str:
        """
            get the attribute name defining this category
        """
        return _category_to_attribute_name[self]


# corresponds to a DFS of the goal structure
_category_to_path = {
    Category.STUDY: tuple(),
    Category.AUTHOR_LIST: (Category.STUDY, ),
    Category.AUTHOR: (Category.STUDY, Category.AUTHOR_LIST),
    Category.AFFILIATION: (Category.STUDY, Category.AUTHOR_LIST, Category.AUTHOR),
    Category.DESCRIPTOR: (Category.STUDY, ),
    Category.KEYWORD: (Category.STUDY, ),
    Category.PUBLICATION_TYPE: (Category.STUDY, ),
    Category.GRANT: (Category.STUDY, )
}

_category_to_attribute_name = {
    Category.STUDY: None,
    Category.AUTHOR_LIST: 'authors',
    Category.AUTHOR: None,
    Category.AFFILIATION: 'affiliations',
    Category.DESCRIPTOR: 'descriptors',
    Category.KEYWORD: 'keywords',
    Category.PUBLICATION_TYPE: 'publication_types',
    Category.GRANT: 'grants',
}


import dateutil.parser as dup

from nbib.exceptions import UnknownTagFormat
from nbib._structure import Category

import re

NESTED_BRACKET_FORMAT = r'^(.*) \[([a-zA-Z\-]+)\]$'
ISSN_FORMAT = r'^([0-9]{4}\-[0-9]{3}[0-9X]) \(([a-zA-Z]+)\)$'


class TagParser:
    def __init__(self, category: Category, attribute: str):
        self._category = category
        self._attribute = attribute

    def get_category(self) -> Category:
        return self._category

    def get_attribute(self) -> str:
        return self._attribute

    def parse(self, lines: list):
        """

        :param lines: the agglomerated set of lines corresponding to the tag being parsed
        :return:
        """
        # it appears that any space needed will already be present on the preceding line
        # note: it may be a bit faster to strip either 1 or 5 characters (leading spaces) by index instead of stripping
        return ''.join(line.lstrip() for line in lines)


class IntParser(TagParser):
    def parse(self, lines: list):
        return int(lines[0])


class DateTimeParser(TagParser):
    """
    this class parses well-formatted, complete dates
    """
    def parse(self, lines: list):
        return dup.parse(lines[0])


class IDParser(TagParser):
    """
    a naughty, stateful parser that depends on the parse() call to be made before the get_attribute() call
    """
    def __init__(self):
        super().__init__(Category.STUDY, 'unspecified_id')

    def parse(self, lines):
        content = ''.join(line.lstrip() for line in lines)
        match = re.match(NESTED_BRACKET_FORMAT, content)

        if not match:
            raise UnknownTagFormat()

        groups = match.groups()
        # technically, we may get name collision this way, but the format is unlikely to change
        # and any additions are unlikely to collide
        self._attribute = groups[1]
        return groups[0]


class ISSNParser(TagParser):
    def __init__(self):
        super().__init__(Category.STUDY, 'unspecified_id')
        self._attribute_prefix = 'unspecified'

    def get_attribute(self) -> str:
        return f'{self._attribute_prefix}_issn'

    def parse(self, lines):
        match = re.match(ISSN_FORMAT, lines[0])

        if not match:
            raise UnknownTagFormat()

        groups = match.groups()
        # technically, we may get name collision this way, but the format is unlikely to change
        # and any additions are unlikely to collide
        self._attribute_prefix = groups[1].lower()
        return groups[0]


class PubMedHistoryParser(TagParser):
    def __init__(self):
        super().__init__(Category.STUDY, 'unspecified_event')

    def parse(self, lines):
        match = re.match(NESTED_BRACKET_FORMAT, lines[0])

        if not match:
            raise UnknownTagFormat()

        groups = match.groups()
        # technically, we may get name collision this way, but the format is unlikely to change
        # and any additions are unlikely to collide
        self._attribute = f"{groups[1]}_time"
        return dup.parse(groups[0])


def get_tag_parsers():
    """
    note: currently, may be stateful, which means a given set of parsers should not be used in parallel

    for a full list: https://pubmed.ncbi.nlm.nih.gov/help/#pubmed-format
    """
    return {
        'AB': TagParser(Category.STUDY, 'abstract'),
        'AD': TagParser(Category.AFFILIATION, 'affiliation'),
        'AID': IDParser(),
        'AU': TagParser(Category.AUTHOR, 'author_abbreviated'),
        # skipping BTI (book title) as irrelevant
        'CI': TagParser(Category.STUDY, 'copyright'),
        # skipping CIN (comment in) since it's information rich + somewhat ambiguous
        'CN': TagParser(Category.STUDY, 'corporate_author'),
        'COIS': TagParser(Category.STUDY, 'conflict_of_interest'),  # note: documentation incorrectly calls this 'COI'
        # skipping CON (comment on)
        # skipping CP (book chapter) as irrelevant
        # skipping CRDT (creation date), since at best is confused with other information
        # skipping CRF (corrected and republished form)
        # skipping CRI (corrected and republished in)
        # skipping CTDT (book contribution date)
        # skipping CTI (collection title)
        # skipping DCOM (nlm completion date) due to near duplication with phst
        # skipping DDIN (dataset described in) no examples found
        # skipping DRIN (dataset use reported in) no examples found
        'DEP': DateTimeParser(Category.STUDY, 'electronic_publication_date'),
        'DP': TagParser(Category.STUDY, 'publication_date'),  # TODO: this is an incomplete date - how to represent?
        # skipping DRDT (book revision date) as irrelevant
        # skipping ECF (expression of concern for) no examples found
        # skipping ECI (expression of concern in) no examples found
        # skipping EDAT (entrez date) duplicated exactly in phst
        # skipping EFR (erratum for) no examples found
        # skipping EIN (erratum in) no examples found
        # skipping ED (book editors) as irrelevant
        # skipping EN (book edition) as irrelevant
        'FAU': TagParser(Category.AUTHOR, 'author'),
        # skipping FED (editors) no examples found
        # skipping FIR (investigator name) no examples found
        # skipping FPS (full personal name as subject) as irrelevant
        # skipping GN (general note) no examples found
        'GR': TagParser(Category.GRANT, 'grant'),
        # skippping GS (gene symbol) as irrelevant
        'IP': TagParser(Category.STUDY, 'journal_issue'),
        # skipping IR (investigator) no examples found
        # skipping IRAD (investigator affiliation) no examples found
        'IS': ISSNParser(),
        # skipping ISBN as irrelevant
        'JID': TagParser(Category.STUDY, 'nlm_journal_id'),  # not an integer because sometimes suffixed with 'R'
        'JT': TagParser(Category.STUDY, 'journal'),
        'LA': TagParser(Category.STUDY, 'language'),
        # skipping LID (location id) as a duplicate of AID
        'LR': DateTimeParser(Category.STUDY, 'last_revision_date'),
        'MH': TagParser(Category.DESCRIPTOR, 'mesh_heading'),
        # skipping MHDA (mesh date) as duplicate of [medline] PHST
        # skipping OAB (other abstract) as irrelevant
        # skipping OABL (other abstract language) as irrelevant
        # skipping OCI (other copyright information) as irrelevant
        # skipping OID (other id) no examples
        # skipping ORI (original report in) no examples
        'OT': TagParser(Category.KEYWORD, 'keyword'),
        # skipping OTO (keyword owner) no examples
        'OWN': TagParser(Category.STUDY, 'citation_owner'),
        # skipping PB as irrelevant
        'PG': TagParser(Category.STUDY, 'pages'),
        'PHST': PubMedHistoryParser(),
        'PL': TagParser(Category.STUDY, 'place_of_publication'),
        # skipping PMCR (pmc released) in favor of exposing the (undocumented) PMC id (if present)
        'PMID': IntParser(Category.STUDY, 'pubmed_id'),
        # skipping PRIN (partial retraction in) no examples
        # skipping PROF (partial retraction of) no examples
        # skipping PS (personal name as subject) no examples
        'PST': TagParser(Category.STUDY, 'publication_status'),
        'PT': TagParser(Category.PUBLICATION_TYPE, 'publication_type'),
        # skipping RF (number of references) no examples
        # skipping RIN (retraction in) no examples
        # skipping RN (EC/RN number) for lack of understanding
        # skipping ROF (retraction of) no examples
        # skipping RPF (republished from) no examples
        # skipping RRI (retracted and republished in) no examples
        # skipping RRF (Retracted and Republished From) no examples
        # skipping SB (Subset) for lack of understanding
        # skipping SFM as no examples / irrelevant
        'SI': TagParser(Category.STUDY, 'secondary_source'),
        # skipping SO (source) as duplicate of a variety of the information already present
        # skipping SPIN (summary for patients in) no examples
        'STAT': TagParser(Category.STUDY, 'nlm_status'),
        'TA': TagParser(Category.STUDY, 'journal_abbreviated'),
        'TI': TagParser(Category.STUDY, 'title'),
        'TT': TagParser(Category.STUDY, 'transliterated_title'),
        # skipping UIN (update in) no examples
        # skipping UOF (update of) no examples
        'VI': TagParser(Category.STUDY, 'journal_volume'),
        # skipping VTI (book volume title) irrelevant
    }



class UnknownTagFormat(Exception):
    """
        occurs when a tag's contents violate expected format
    """
    pass


class MalformedLine(Exception):
    """
        occurs when an nbib line doesn't conform to the standard {Tag|spaces}-value format
    """
    pass



import re
from typing import Dict, List

from nbib.exceptions import MalformedLine, UnknownTagFormat
from nbib._tags import get_tag_parsers

_LINE_REGEX = r'^([ A-Z]{4})(-| ) (.*)$'


def read_file(path: str):
    """
        Read nbib formated references from a file

        :param path: an nbib formatted file
        :return: a list of parsed references. references contain a variety of attributes structured according to the _tags and _structure submodules

        :raises MalformedLine: when a line does match required structure
        :raises UnknownTagFormat: when a tag value with structured data doesn't follow expected format
    """
    with open(path, mode='r') as f:
        return read(f.read())


def read(content: str) -> List[Dict]:
    """
        Read nbib formated references from a string

        :param content: an nbib formatted string
        :return: a list of parsed references. references contain a variety of attributes structured according to the _tags and _structure submodules

        :raises MalformedLine: when a line does match required structure
        :raises UnknownTagFormat: when a tag value with structured data doesn't follow expected format
    """
    tag_parsers = get_tag_parsers()
    fp_refs = _first_pass_parse(content, tag_parsers)
    sp_refs = _second_pass_parse(fp_refs)

    return sp_refs


def _parse_and_place_tag(current_tag: str, current_tag_content: list, current_ref: dict, tag_parsers: dict):
    parser = tag_parsers.get(current_tag, None)
    # an exception may be right in some cases, but our tag mapping currently isn't broad enough to except
    if not parser:
        return

    ##############
    # parse the incoming tag
    ##############

    parsed_content = parser.parse(current_tag_content)
    attribute_name = parser.get_attribute()

    category = parser.get_category()

    ##############
    # advance in the ref to the desired write location, creating objects along the way as needed
    ##############
    ref_write_location = current_ref
    category_path = category.get_category_path()
    if len(category_path) > 0:
        parent_category = category_path[0]
        if len(category_path) > 1:
            navigation = category_path[1:] + (category, )
        else:
            navigation = (category, )

        for ix, cat in enumerate(navigation):
            if parent_category.list_appendable():
                # case: the leaf category is for sequential tags (i.e. attribute appendable rooted beneath list
                # appendable) and we've reached a new start tag. create a new entry
                if ix == len(navigation) - 1:
                    if cat.attribute_appendable() and cat.get_start_tag() == current_tag:
                        ref_write_location.append({})

                # to navigate the list, we simply pick the last (most recent) element
                # if the list is empty, a tag is unexpectedly located. this can happen (e.g. affiliation being used for
                # an entire study & for individual authors). we can either throw or drop it :/
                if len(ref_write_location) == 0:
                    return
                ref_write_location = ref_write_location[-1]
            # to navigate the dict, we pick out the correct key (attribute)
            elif parent_category.attribute_appendable():
                cat_attribute = cat.get_attribute_name()
                if cat_attribute not in ref_write_location:
                    if cat.list_appendable():
                        ref_write_location[cat_attribute] = []
                    elif cat.attribute_appendable():
                        ref_write_location[cat_attribute] = {}
                    else:
                        raise ValueError('Categories must be either list or attribute appendable!')

                ref_write_location = ref_write_location[cat_attribute]
            else:
                raise ValueError('Categories must be either list or attribute appendable!')
            parent_category = cat

    ##############
    # perform the write
    ##############
    if category.list_appendable():
        ref_write_location.append(parsed_content)
    elif category.attribute_appendable():
        ref_write_location[attribute_name] = parsed_content
    else:
        raise ValueError('Categories must be either list or attribute appendable!')


def _parse_line(line: str) -> tuple:
    """
        returns tag and tag content, in order
        if the line is a continuation, tag will be None
    """
    match = re.match(_LINE_REGEX, line)
    if match:
        groups = match.groups()
        if groups[1] == '-':
            return groups[0].rstrip(), groups[2]
        else:
            return None, groups[2]

    raise MalformedLine()


def _first_pass_parse(content: str, tag_parsers: dict) -> list:
    """
        in first pass parsing, every tag is mapped to a single attribute of a ref
    """
    lines = content.split('\n')

    refs = []

    current_ref = {}
    # tags are parsed only when the next tag (or end of record) is encountered (to account for multi-line tags)
    # so we keep a running collection of all lines in the tag
    current_tag = None
    current_tag_content = []

    for line in lines:
        if line == '':  # end of record
            _parse_and_place_tag(current_tag, current_tag_content, current_ref, tag_parsers)
            # filter end of file (which technically starts a new entry with two newlines) and otherwise empty refs
            if not current_ref == {}:
                refs.append(current_ref)
                current_ref = {}
            current_tag_content = []
            current_tag = None
        else:
            next_tag, next_tag_content = _parse_line(line)

            if next_tag:
                _parse_and_place_tag(current_tag, current_tag_content, current_ref, tag_parsers)
                current_tag = next_tag
                current_tag_content = [next_tag_content]
            else:
                current_tag_content.append(next_tag_content)

    return refs


def _second_pass_parse(refs: list) -> list:
    """
        in second pass parsing, parsed attributes are exploded to new attributes
        these are currently hardcodes that depend on some characteristics of the structure module
    """
    for ref in refs:
        # expand MeSH heading strings into descriptor, qualifier (optional) pairs
        if 'descriptors' in ref:
            new_descriptors = []
            for raw_descriptor in ref['descriptors']:
                headings = [h for h in raw_descriptor.split('/')]
                descriptor = headings[0]
                if len(headings) > 1:
                    for qualifier in headings[1:]:
                        new_descriptors.append({
                            'descriptor': descriptor.lstrip('*'),
                            'qualifier': qualifier.lstrip('*'),
                            'major': qualifier.startswith('*') or descriptor.startswith('*')
                        })
                else:
                    new_descriptors.append({'descriptor': descriptor.lstrip('*'), 'major': descriptor.startswith('*')})
            ref['descriptors'] = new_descriptors

        if 'authors' in ref:
            for auth in ref['authors']:
                name = auth['author']
                name_parts = [n.strip() for n in name.split(',')]
                # it looks like PubMed almost always spits out a last, first pair.
                # if that structure is violated, just skip it
                if len(name_parts) == 2:
                    # leave the rest of the name attributes in :/
                    auth['first_name'] = name_parts[1]
                    auth['last_name'] = name_parts[0]

    return refs
